from collections import defaultdict
import random

from .core import TickProcess

class PacketGenerator(TickProcess):
    def __init__(self, ticks_per_packet=1, start_delay=0, time_to_live=60):
        super().__init__(ticks_per_packet)

        if time_to_live is None:
            raise ValueError('time_to_live must be specified')

        self.start_delay = start_delay
        self.time_to_live = time_to_live
        self.packets = []

    def process(self, **kwargs):
        network = kwargs['network']
        packet_id = 0
        def create(packet_id):
            source, dest = random.choice(network.links)
            packet = Packet(packet_id, source, dest,
                            self.env.now + self.time_to_live, None)
            self.packets.append(packet)
            source.recv(packet)
        yield self.env.timeout(self.start_delay)
        print('starting packets')
        while True:
            create(packet_id)
            yield self.tick()
            packet_id += 1

    @property
    def stats(self):
        stats = {}
        for packet in self.packets:
            for stat, value in packet.stats.items():
                if stat not in stats:
                    stats[stat] = value
                else:
                    stats[stat] += value
        stats['packets'] = len(self.packets)
        stats['recieved'] = len([p for p in self.packets if p.recieved])
        stats['delivery_ratio'] = stats['recieved'] / stats['packets']
        stats['duplicates'] = sum([p.duplicates for p in self.packets])
        stats['transmissions'] = sum([p.sent for p in self.packets])
        stats['delivery_cost'] = stats['transmissions'] / stats['recieved']
        return stats

    def __str__(self):
        ret = ''
        for stat, value in sorted(self.stats.items()):
            ret += f'{stat}: {value}\n'
        return ret


class Packet:
    def __init__(self, packet_id, source, destination, ttl, payload):
        self.id = packet_id
        self.source = source
        self.destination = destination
        self.ttl = ttl
        self.payload = payload
        self.size = 0

        self.path = []

        self.stats = defaultdict(int)

        self.sent = 0
        self.recieved = False
        self.duplicates = 0
        self.dropped = 0

    def expired(self, now):
        return self.ttl < now

    def send(self, sender, targets, reason=None):
        target_ids = [ n.id for n in targets ]
        if reason is None:
            self.path.append((sender.id, target_ids))
        else:
            self.stats[f'send_{reason}'] += 1
            self.path.append((sender.id, target_ids, reason))
        self.sent += 1

    def recv(self):
        if self.recieved:
            self.duplicates += 1
        self.recieved = True

    def drop(self):
        self.dropped += 1

    def __str__(self):
        return "Packet(id={}, src={}, dst={})".format(
            self.id,
            self.source,
            self.destination
        )


